"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const GraphCluster_1 = require("../../graph/GraphCluster");
const GraphGenerator_1 = require("../../graph/GraphGenerator");
const GraphNode_1 = require("../../graph/GraphNode");
const GraphOptimizer_1 = require("../../graph/utils/GraphOptimizer");
const GraphVisitor_1 = require("../../graph/utils/GraphVisitor");
const DateHelper_1 = require("../../utils/DateHelper");
const DocGeneratorBase_1 = require("../shared/DocGeneratorBase");
const DotStyleGenerator_1 = require("./utils/DotStyleGenerator");
const MapNameToId_1 = require("./utils/MapNameToId");
class DotDocGenerator extends DocGeneratorBase_1.DocGeneratorBase {
    constructor(config, outputter) {
        super(config, outputter);
        this.config = config;
        this.outputter = outputter;
        this.mapNameToId = new MapNameToId_1.MapNameToId();
        this.styler = new DotStyleGenerator_1.DotStyleGenerator(config, outputter);
    }
    generateDoc(packageConfig) {
        const graph = this.generateGraph(packageConfig);
        this.outputSectionSeparator("Header");
        this.outputHeader();
        this.outputter.outputLine(``);
        this.outputSectionSeparator("Styling");
        this.styler.outputStyling();
        this.outputter.outputLine(``);
        this.outputSectionSeparator("Nodes and Clusters");
        graph.nodes.forEach(node => {
            if (node instanceof GraphCluster_1.GraphCluster) {
                this.outputCluster(node);
                return;
            }
            if (node instanceof GraphNode_1.GraphNode) {
                this.outputGraphNode(node);
            }
        });
        this.outputter.outputLine(``);
        this.outputSectionSeparator("Edges");
        this.outputEdges(graph);
        this.outputter.outputLine(``);
        this.outputFooter();
    }
    generateGraph(packageConfig) {
        const generator = new GraphGenerator_1.GraphGenerator(this.config, this.filter);
        const graph = generator.generateGraph(packageConfig);
        if (this.config.dot.isGraphOptimizerEnabled) {
            const optimizer = new GraphOptimizer_1.GraphOptimizer();
            optimizer.optimize(graph);
        }
        return graph;
    }
    outputHeader() {
        this.outputter.outputLine(`/* auto-generated by tslint-folders docs tool at ${DateHelper_1.DateHelper.nowHumanReadable()}*/`);
        this.outputter.outputLine("digraph packages {");
        this.outputter.increaseIndent();
        this.outputGraphSettings();
    }
    outputGraphSettings() {
        this.outputSectionSeparator("Graph settings");
        this.outputter.outputLine(`graph [`);
        this.outputter.increaseIndent();
        this.outputter.outputLines([
            `// needed to allow edge to have *cluster* as a destination:`,
            `compound=true`,
            // NOT using concentrate, as tends to combine edges from within a cluster, with edges from the cluster,
            // so user cannot see which edges are only from the cluster.
            `label = "${this.config.dot.title}"`,
            `labelloc = t`,
            ``,
            `//dpi = 200`,
            `ranksep=0.65`,
            `nodesep=0.40`,
            // top to bottom:
            `rankdir=TB`,
            ``,
            `style="filled"`,
            ``,
            `len=0`
        ]);
        this.outputter.decreaseIndent();
        this.outputter.outputLine(`]`);
    }
    outputFooter() {
        this.outputter.decreaseIndent();
        this.outputter.outputLine("}");
    }
    outputSectionSeparator(sectionName) {
        this.outputter.outputLine(`/* ${sectionName} ================================= */`);
    }
    outputGraphNode(node, prefix) {
        this.outputScopeBegin();
        this.styler.outputGraphNodeStyle(node.nodeType);
        this.outputNode(node, prefix);
        this.outputScopeEnd();
        this.outputter.outputLine("");
    }
    outputScopeBegin() {
        this.outputter.outputLine("{");
    }
    outputScopeEnd() {
        this.outputter.outputLine("}");
    }
    /**
     * Output the separate graph cluster.
     */
    outputCluster(cluster) {
        const isContainer = cluster.nodes.length > 0;
        if (!isContainer) {
            return;
        }
        this.outputContainerNodeStart(cluster);
        this.outputSubNodes(cluster);
        this.outputContainerNodeEnd();
        this.outputter.outputLine(``);
    }
    outputContainerNodeStart(cluster) {
        this.outputter.outputLine(`subgraph cluster_${cluster.id} {`);
        this.outputter.increaseIndent();
        this.styler.outputContainerStyle(cluster.clusterType);
        const formattedDescription = cluster.description.length > 0 ? ` - ${cluster.description}` : "";
        this.outputter.outputLine(`label = "${cluster.name}${formattedDescription}"`);
    }
    outputContainerNodeEnd() {
        // styling at END so is not broken by contained node styling
        this.outputter.decreaseIndent();
        this.outputter.outputLine("}");
    }
    outputNode(node, prefix) {
        const packageIdKey = this.getPrefixedPackageId(node.id, prefix);
        const fillColor = this.getColorNumber(packageIdKey);
        this.outputter.outputLine(`${node.id} [label="${node.name}\n${node.description}" fillcolor=${fillColor}]`);
    }
    getColorNumber(packageIdKey) {
        // TODO xxx review
        if (!this.mapNameToId.hasId(packageIdKey)) {
            this.mapNameToId.getId(packageIdKey);
        }
        let packageNumber = this.mapNameToId.getNumberOrThrow(packageIdKey);
        if (packageNumber > this.config.dot.maxColors) {
            packageNumber = packageNumber % this.config.dot.maxColors;
        }
        if (packageNumber === 0) {
            // for a graphviz color scheme, the color number must be 1 based:
            packageNumber = 1;
        }
        return packageNumber;
    }
    getPrefixedPackageId(packageName, prefix) {
        return `${prefix}_${packageName}`;
    }
    outputEdges(root) {
        const visitor = new GraphVisitor_1.GraphVisitor(root);
        visitor.visitEdges(edge => this.outputEdge(edge));
    }
    outputEdge(edge) {
        let destination = edge.destination;
        let origin = edge.origin;
        // set a dummy label, to make edges further apart.
        // this is to avoid edges from within a cluster, seeming to have same origin
        let extraAttribute = `label=" " `;
        if (edge.destination instanceof GraphCluster_1.GraphCluster) {
            // the destination needs to switch to some node in the cluster:
            const nodes = edge.destination.nodes;
            if (nodes.length === 0) {
                console.warn(`// warning: The cluster ${edge.destination.id} is empty, but is the destination for an edge.`);
                return;
            }
            destination = nodes[0];
            extraAttribute += ` lhead=cluster_${edge.destination.id}`;
        }
        if (edge.origin instanceof GraphCluster_1.GraphCluster) {
            const nodes = edge.origin.nodes;
            if (nodes.length === 0) {
                console.warn(`// warning: The cluster ${edge.origin.id} is empty, but is the origin for an edge.`);
                return;
            }
            origin = nodes[0];
            extraAttribute += ` ltail=cluster_${edge.origin.id}`;
        }
        this.outputter.outputLine(`${origin.id}-> ${destination.id} [${extraAttribute}]`);
    }
    outputSubNodes(cluster) {
        if (cluster.nodes.length === 0) {
            this.outputter.outputLine("");
            return;
        }
        cluster.nodes.forEach(node => {
            if (node instanceof GraphCluster_1.GraphCluster) {
                this.outputCluster(node);
                return;
            }
            this.outputGraphNode(node, cluster.name);
            this.outputter.outputLine("");
        });
    }
}
exports.DotDocGenerator = DotDocGenerator;
//# sourceMappingURL=DotDocGenerator.js.map