"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const path = require("path");
const Lint = require("tslint");
const ConfigFactory_1 = require("./config/ConfigFactory");
const FilenamesRuleConfig_1 = require("./model/FilenamesRuleConfig");
const EnumUtils_1 = require("./utils/EnumUtils");
const FilenameCasingUtils_1 = require("./utils/FilenameCasingUtils");
const ImportRuleUtils_1 = require("./utils/ImportRuleUtils");
exports.FILE_NAMES_RULE_ID = "tsf-folders-file-names";
/** Custom version of the standard file-name-casing rule, that allows for *multiple* casings.
 * ref: https://github.com/palantir/tslint/blob/master/src/rules/fileNameCasingRule.ts
 */
class Rule extends Lint.Rules.AbstractRule {
    static FAILURE_STRING(expectedCasings) {
        const stylizedCasings = expectedCasings
            .map(casing => this.stylizedNameForCasing(casing))
            .join(" or ");
        // include the rule ID, to make it easier to disable
        return `File name must be ${stylizedCasings} (${exports.FILE_NAMES_RULE_ID})`;
    }
    static stylizedNameForCasing(casing) {
        switch (casing) {
            case FilenamesRuleConfig_1.Casing.CamelCase:
                return "camelCase";
            case FilenamesRuleConfig_1.Casing.PascalCase:
                return "PascalCase";
            case FilenamesRuleConfig_1.Casing.KebabCase:
                return "kebab-case";
            case FilenamesRuleConfig_1.Casing.SnakeCase:
                return "snake_case";
            default:
                throw new Error(`unhandled casing ${casing}`);
        }
    }
    apply(sourceFile) {
        const config = ConfigFactory_1.ConfigFactory.createForFilenames(this.getOptions());
        if (ImportRuleUtils_1.ImportRuleUtils.shouldIgnorePath(sourceFile.fileName, config.ignorePaths)) {
            return [];
        }
        const parsedPath = path.parse(sourceFile.fileName);
        const filename = parsedPath.name;
        const isAnAllowedCasing = config.casings.some(casingName => {
            const casing = EnumUtils_1.EnumUtils.parseCasing(casingName);
            const isAllowed = FilenameCasingUtils_1.FilenameCasingUtils.isCased(filename, casing);
            return isAllowed;
        });
        if (!isAnAllowedCasing) {
            return [
                new Lint.RuleFailure(sourceFile, 0, 0, Rule.FAILURE_STRING(config.casings), exports.FILE_NAMES_RULE_ID)
            ];
        }
        return [];
    }
}
exports.Rule = Rule;
//# sourceMappingURL=tsfFoldersFileNamesRule.js.map