"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const GeneralRuleUtils_1 = require("./GeneralRuleUtils");
const PackageConfigHelper_1 = require("./PackageConfigHelper");
var PathSource;
(function (PathSource) {
    PathSource[PathSource["SourceFilePath"] = 0] = "SourceFilePath";
    PathSource[PathSource["ImportText"] = 1] = "ImportText";
})(PathSource = exports.PathSource || (exports.PathSource = {}));
var ImportRuleUtils;
(function (ImportRuleUtils) {
    ImportRuleUtils.IS_DEBUG_ENABLED = false;
    function determinePackageLocationFromPath(filePath, ruleId, config, pathSource, thisPackageLocation) {
        const dirs = cleanPath(filePath).split("/");
        if (ImportRuleUtils.IS_DEBUG_ENABLED) {
            console.log(`\nchecking ${PathSource[pathSource]} against path:`, dirs.join(","));
        }
        let packageName = determinePackageName(config, dirs, pathSource);
        if (packageName === null ||
            (pathSource === PathSource.SourceFilePath &&
                !PackageConfigHelper_1.PackageConfigHelper.hasPackage(config, packageName))) {
            return {
                packageName: filePath
            };
        }
        const packageFolderAndName = determinePackageFolderAndName(config, dirs, filePath, thisPackageLocation, packageName, pathSource, ruleId);
        packageName = packageFolderAndName.packageName;
        if (!packageFolderAndName.packageFolder) {
            return {
                packageName: packageName
            };
        }
        if (packageFolderAndName.packageFolder.subFolders.length === 0) {
            return {
                packageName: packageName,
                packageFolder: packageFolderAndName.packageFolder
            };
        }
        return determinePackageLocationWithSubFolder(dirs, packageName, packageFolderAndName.packageFolder, pathSource);
    }
    ImportRuleUtils.determinePackageLocationFromPath = determinePackageLocationFromPath;
    // TODO xxx review - too many params!
    function determinePackageFolderAndName(config, dirs, filePath, thisPackageLocation, packageName, pathSource, ruleId) {
        let activePackageName = packageName;
        let packageFolder;
        if (pathSource === PathSource.ImportText && isRelativeImport(dirs[0])) {
            if (!thisPackageLocation) {
                throw new Error("for path source = ImportText - expect thisPackageLocation to be set");
            }
            // assumption: we are importing from 'this' package
            // (assuming is not ..'ing back into other package! - that would require 'virtual directories')
            packageFolder = thisPackageLocation.packageFolder;
        }
        else {
            try {
                packageFolder = PackageConfigHelper_1.PackageConfigHelper.getPackage(config, packageName);
            }
            catch (error) {
                GeneralRuleUtils_1.GeneralRuleUtils.buildFailureString(`unexpected: import had a recognised package: [${packageName}] but could not find the PackageFolder in the config`, ruleId);
                packageFolder = undefined;
                activePackageName = filePath;
            }
        }
        return {
            packageFolder: packageFolder,
            packageName: activePackageName
        };
    }
    function determinePackageName(config, dirs, pathSource) {
        let packageName = null;
        switch (pathSource) {
            case PathSource.ImportText:
                {
                    // take the 1st part of the path:
                    // (ignore local directories that happen to have same name as a package)
                    packageName = dirs[0];
                }
                break;
            case PathSource.SourceFilePath: {
                dirs.forEach(dir => {
                    if (PackageConfigHelper_1.PackageConfigHelper.hasPackage(config, dir)) {
                        if (packageName === null) {
                            // take the 1st recognised folder:
                            packageName = dir;
                        }
                        else if (ImportRuleUtils.IS_DEBUG_ENABLED) {
                            // this can occur with package names like 'utils'
                            console.warn(`import has more than one recognised package: [${packageName},${dir}]`);
                        }
                    }
                });
                break;
            }
            default: {
                throw new Error(`unhandled PathSource ${pathSource}`);
            }
        }
        return packageName;
    }
    function determinePackageLocationWithSubFolder(dirs, packageName, packageFolder, pathSource) {
        let packageSubFolder;
        let activePackageName = packageName;
        switch (pathSource) {
            case PathSource.ImportText:
                {
                    // take the 1st part of the path:
                    // (ignore local directories that happen to have same name as a package)
                    activePackageName = dirs[0];
                    const isImportingFromSubFolder = isRelativeImport(dirs[0]);
                    if (isImportingFromSubFolder) {
                        for (let i = 1; i < dirs.length; i++) {
                            const subFolder = packageFolder.subFolders.find(folder => folder.importPath === dirs[i]);
                            if (subFolder) {
                                packageSubFolder = subFolder;
                                activePackageName = subFolder.importPath;
                                break;
                            }
                        }
                    }
                }
                break;
            case PathSource.SourceFilePath: {
                let hasPackageDir = false;
                dirs.forEach(dir => {
                    if (packageSubFolder) {
                        return;
                    }
                    if (packageFolder.importPath === dir) {
                        hasPackageDir = true;
                        return;
                    }
                    {
                        if (hasPackageDir) {
                            const subFolder = packageFolder.subFolders.find(folder => folder.importPath === dir);
                            if (subFolder) {
                                packageSubFolder = subFolder;
                                return;
                            }
                        }
                    }
                });
                break;
            }
            default: {
                throw new Error(`unhandled PathSource ${pathSource}`);
            }
        }
        return {
            packageName: activePackageName,
            packageFolder: packageFolder,
            packageSubFolder: packageSubFolder
        };
    }
    function isRelativeImport(pathDir) {
        return pathDir === "." || pathDir === "..";
    }
    ImportRuleUtils.isRelativeImport = isRelativeImport;
    function cleanPath(filePath) {
        let cleaned = filePath.trim();
        cleaned = cleaned.replace(/['"]+/g, "");
        cleaned = cleaned.replace(/[\\]+/g, "\\");
        return cleaned;
    }
    function isThisPackageThirdParty(thisFolderLocation, node) {
        if (isPackageThirdParty(thisFolderLocation)) {
            if (ImportRuleUtils.IS_DEBUG_ENABLED) {
                console.log(node.getSourceFile().fileName, "- this package ThirdParty!", thisFolderLocation.packageName);
            }
            return true;
        }
        return false;
    }
    ImportRuleUtils.isThisPackageThirdParty = isThisPackageThirdParty;
    function isPackageThirdParty(folderLocation) {
        return !folderLocation.packageFolder && !folderLocation.packageSubFolder;
    }
    ImportRuleUtils.isPackageThirdParty = isPackageThirdParty;
    function logPackageAndImport(node, thisPackageLocation, importPackageLocation) {
        if (ImportRuleUtils.IS_DEBUG_ENABLED) {
            console.log(node.getSourceFile().fileName, "- this package:", thisPackageLocation.packageName, getPackageName(thisPackageLocation.packageFolder), "- import:", importPackageLocation.packageName, getPackageName(importPackageLocation.packageFolder));
        }
    }
    ImportRuleUtils.logPackageAndImport = logPackageAndImport;
    function getPackageName(packageFolder) {
        return packageFolder ? packageFolder.importPath : "(unknown)";
    }
    function shouldIgnoreFile(node, ignorePaths) {
        const filePath = node.getSourceFile().fileName;
        return shouldIgnorePath(filePath, ignorePaths);
    }
    ImportRuleUtils.shouldIgnoreFile = shouldIgnoreFile;
    function shouldIgnorePath(path, ignorePaths) {
        return ignorePaths.some(ignore => path.indexOf(ignore) >= 0);
    }
    ImportRuleUtils.shouldIgnorePath = shouldIgnorePath;
})(ImportRuleUtils = exports.ImportRuleUtils || (exports.ImportRuleUtils = {}));
//# sourceMappingURL=ImportRuleUtils.js.map